import socket

import dns.query
import dns.resolver
import dns.zone


# Checks for zone transfers on that domain. Very rare to succeed, but when you
# do, it's a gold mine
def dns_zonetransfer(self, target):
  print("Testing for zone transfers")

  zonetransfers = []
  resolver = dns.resolver.Resolver()

  try:
    answers = resolver.query(target, 'NS')
  except Exception as e:
    self.handle_exception(e, "Error checking for Zone Transfers")
    return

  resolved_ips = []

  for ns in answers:
    ns = str(ns).rstrip('.')
    resolved_ips.append(socket.gethostbyname(ns))

  for ip in resolved_ips:
    try:
      zone = dns.zone.from_xfr(dns.query.xfr(ip, target))
      for name, node in zone.nodes.items():
        name = str(name)
        if name not in ["@", "*"]:
          zonetransfers.append(name + '.' + target)
    except:
      pass

  if zonetransfers:
    print("\tZone transfers possible:")
    for zone in zonetransfers:
      print(zone)



"""The base command."""


class Base(object):
  """A base command."""

  def __init__(self, options, *args, **kwargs):
    self.options = options
    self.args = args
    self.kwargs = kwargs

  def run(self):
    raise NotImplementedError(
      'run() method must be implemented by the overloading class')



import re

import requests


def search_netcraft(self, target):
  print("Searching NetCraft.com")
  headers = {'Pragma':          'no-cache', 'DNT': '1',
             'Accept-Encoding': 'gzip, deflate, br',
             'Accept-Language': 'en-US,en;q=0.9,it;q=0.8',
             'User-Agent':      'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_13_2) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/62.0.3202.94 Safari/537.36',
             'Accept':          'text/html,application/xhtml+xml,application/xml;q=0.9,image/webp,image/apng,*/*;q=0.8',
             'Cache-Control':   'no-cache',
             'Referer':         'https://searchdns.netcraft.com/?restriction=site+ends+with&host=',
             'Connection':      'keep-alive', }

  params = (('restriction', 'site contains'), ('host', target))
  try:
    res = requests.get('https://searchdns.netcraft.com/', headers=headers,
                       params=params)
    scraped = res.text
    trimmed = scraped[scraped.find('<div class="blogtitle">'):scraped.rfind(
      '<div id="copyright">')]
    subdomain_finder = re.compile(
      '<a href="http://toolbar.netcraft.com/site_report\?url=(.*)">')
    links = subdomain_finder.findall(trimmed)
    for domain in links:
      if domain.strip() not in self.domains and domain.endswith("." + target):
        self.domains.append(domain.strip())
        if self.options["--verbose"]:
          print("Netcraft Found Domain:", domain.strip())
  except Exception as e:
    self.handle_exception(e, "Error searching NetCraft")
    pass



import socket
from json import dumps

import shodan

from anubis.utils.ColorPrint import ColorPrint


def search_shodan(self):
  print("Searching Shodan.io for additional information")
  try:
    from anubis.API import SHODAN_KEY
  except ImportError:
    ColorPrint.red("Unable to import API keys - make sure API.py exists!")
    return

  api = shodan.Shodan(SHODAN_KEY)
  for i in range(len(self.options["TARGET"])):
    try:
      results = api.host(socket.gethostbyname(self.options["TARGET"][i]))

      if self.options["--verbose"]:
        print(dumps(results, indent=2, sort_keys=True))

      print('Server Location: ' + str(results['city']) + ", " + str(
        results['country_code']) + ' - ' + str(results['postal_code']))

      print("ISP  or Hosting Company: %s" % str(results['isp']))

      if results['os'] is not None:
        print("Possible OS: %s" % str(results['os']))
    except Exception as e:
      self.handle_exception(e, "Error retrieving additional info")



"""The target command."""
import queue
import signal
import sys
import threading
from io import StringIO

from anubis.utils.search_worker import SearchWorker
from anubis.utils.signal_handler import SignalHandler


def recursive_search(self):
  print("Starting recursive search - warning, might take a long time")
  domains = self.clean_domains(self.domains)
  domains_unique = set(domains)
  num_workers = 10

  if self.options["--queue-workers"]:
    num_workers = int(self.options["--queue-workers"])

  stopper = threading.Event()
  url_queue = queue.Queue()
  for domain in domains_unique:
    url_queue.put(domain)

  # we need to keep track of the workers but not start them yet
  workers = [SearchWorker(url_queue, self.domains, stopper, self) for _ in
             range(num_workers)]

  # create our signal handler and connect it
  handler = SignalHandler(stopper, workers)
  signal.signal(signal.SIGINT, handler)

  if not self.options["--verbose"]:
    # catch stdout and replace it with our own
    self.stdout, sys.stdout = sys.stdout, StringIO()

  # start the threads!
  for worker in workers:
    worker.start()

  # wait for the queue to empty
  url_queue.join()

  sys.stdout = self.stdout



from json import dumps, loads

import requests

from anubis.utils.ColorPrint import ColorPrint


def search_anubisdb(self, target):
  print("Searching Anubis-DB")
  res = requests.get("https://jldc.me/anubis/subdomains/" + target)
  if res.status_code == 200 and res.text:
    subdomains = loads(res.text)
    for subdomain in subdomains:
      if subdomain not in self.domains:
        self.domains.append(subdomain)


def send_to_anubisdb(self, target):
  if len(target) == 1:
    print("Sending to AnubisDB")
    data = {'subdomains': dumps(self.domains)}
    # Sends found subdomains to Anubis (max 10,000/post)
    res = requests.post("https://jldc.me/anubis/subdomains/" + target[0],
                        data=data)
    if res.status_code != 200:
      ColorPrint.red("Error sending results to AnubisDB - Status Code: " + str(
        res.status_code))
  else:
    print("Cannot send multiple domains to AnubisDB")



from cryptography import x509
from cryptography.x509.oid import ExtensionOID
from sslyze import ScanCommand, Scanner, ServerConnectivityTester, ServerNetworkLocationViaDirectConnection, ServerScanRequest
from sslyze.errors import ConnectionToServerFailed


def search_subject_alt_name(self, target):
  print("Searching for Subject Alt Names")
  try:
    server_location = ServerNetworkLocationViaDirectConnection.with_ip_address_lookup(
      target, 443)

    # Do connectivity testing to ensure SSLyze is able to connect
    try:
      server_info = ServerConnectivityTester().perform(server_location)
    except ConnectionToServerFailed as e:
      # Could not connect to the server; abort
      print(f"Error connecting to {server_location}: {e.error_message}")
      return

    # Then queue some scan commands for the server
    scanner = Scanner()
    server_scan_req = ServerScanRequest(server_info=server_info, scan_commands={
      ScanCommand.CERTIFICATE_INFO}, )
    scanner.queue_scan(server_scan_req)
    # Then retrieve the results
    for server_scan_result in scanner.get_results():
      # Certificate info results
      certinfo_result = server_scan_result.scan_commands_results[
        ScanCommand.CERTIFICATE_INFO]

      # Direct object reference is pretty bad, but then again so is the crypto.x509 object implementation, so...
      cert_deployment = certinfo_result.certificate_deployments[0]
      chain = cert_deployment.received_certificate_chain[0]
      ext = chain.extensions.get_extension_for_oid(
        ExtensionOID.SUBJECT_ALTERNATIVE_NAME)
      for entry in ext.value.get_values_for_type(x509.DNSName):
        if entry.strip() not in self.domains:
          self.domains.append(entry.strip())

  except Exception as e:
    self.handle_exception(e)


def ssl_scan(self, target):
  print("Running SSL Scan")
  # Define the server that you want to scan
  server_location = ServerNetworkLocationViaDirectConnection.with_ip_address_lookup(
    target, 443)

  try:
    # Do connectivity testing to ensure SSLyze is able to connect
    try:
      server_info = ServerConnectivityTester().perform(server_location)
    except ConnectionToServerFailed as e:
      # Could not connect to the server; abort
      print(f"Error connecting to {server_location}: {e.error_message}")
      return

    # Then queue some scan commands for the server
    scanner = Scanner()
    server_scan_req = ServerScanRequest(server_info=server_info, scan_commands={
      ScanCommand.CERTIFICATE_INFO, ScanCommand.SSL_2_0_CIPHER_SUITES,
      ScanCommand.TLS_1_0_CIPHER_SUITES, ScanCommand.TLS_1_1_CIPHER_SUITES,
      ScanCommand.TLS_1_2_CIPHER_SUITES, ScanCommand.TLS_1_3_CIPHER_SUITES,
      ScanCommand.HEARTBLEED, ScanCommand.HTTP_HEADERS}, )
    scanner.queue_scan(server_scan_req)

    # Then retrieve the results
    for server_scan_result in scanner.get_results():
      print(
        f"\nResults for {server_scan_result.server_info.server_location.hostname}:")

      heartbleed_vuln = server_scan_result.scan_commands_results[
        ScanCommand.HEARTBLEED].is_vulnerable_to_heartbleed
      print(f"\nIs vulnerable to heartbleed? {heartbleed_vuln}")

      print("\nAccepted cipher suites for TLS 1.0:")
      for accepted_cipher_suite in server_scan_result.scan_commands_results[
        ScanCommand.TLS_1_0_CIPHER_SUITES].accepted_cipher_suites:
        print(f"* {accepted_cipher_suite.cipher_suite.name}")

      print("\nAccepted cipher suites for TLS 1.1:")
      for accepted_cipher_suite in server_scan_result.scan_commands_results[
        ScanCommand.TLS_1_1_CIPHER_SUITES].accepted_cipher_suites:
        print(f"* {accepted_cipher_suite.cipher_suite.name}")

      print("\nAccepted cipher suites for TLS 1.2:")
      for accepted_cipher_suite in server_scan_result.scan_commands_results[
        ScanCommand.TLS_1_2_CIPHER_SUITES].accepted_cipher_suites:
        print(f"* {accepted_cipher_suite.cipher_suite.name}")

      print("\nAccepted cipher suites for TLS 1.3:")
      for accepted_cipher_suite in server_scan_result.scan_commands_results[
        ScanCommand.TLS_1_3_CIPHER_SUITES].accepted_cipher_suites:
        print(f"* {accepted_cipher_suite.cipher_suite.name}")

      # SSL 2.0 results
      ssl2_result = server_scan_result.scan_commands_results[
        ScanCommand.SSL_2_0_CIPHER_SUITES]
      print("\nAccepted cipher suites for SSL 2.0:")
      for accepted_cipher_suite in ssl2_result.accepted_cipher_suites:
        print(f"* {accepted_cipher_suite.cipher_suite.name}")

      # Certificate info results
      certinfo_result = server_scan_result.scan_commands_results[
        ScanCommand.CERTIFICATE_INFO]
      print("\nCertificate info:")
      for cert_deployment in certinfo_result.certificate_deployments:
        print(
          f"Leaf certificate: \n{cert_deployment.received_certificate_chain_as_pem[0]}")

  except Exception as e:
    self.handle_exception(e, "Error running SSL scan")
    pass



"""The target command."""
import sys

import re
import socket
import threading
from urllib.parse import urlsplit

from anubis.scanners.anubis_db import search_anubisdb, send_to_anubisdb
from anubis.scanners.crt import search_crtsh
from anubis.scanners.dnsdumpster import search_dnsdumpster
from anubis.scanners.dnssec import dnssecc_subdomain_enum
from anubis.scanners.hackertarget import subdomain_hackertarget
from anubis.scanners.netcraft import search_netcraft
from anubis.scanners.nmap import scan_host
from anubis.scanners.recursive import recursive_search
from anubis.scanners.shodan import search_shodan
from anubis.scanners.ssl import search_subject_alt_name, ssl_scan
from anubis.scanners.virustotal import search_virustotal
from anubis.scanners.zonetransfer import dns_zonetransfer
from anubis.utils.ColorPrint import ColorPrint
from .base import Base


class Target(Base):
  """Main enumeration module"""
  domains = list()
  ip = str()
  dedupe = set()

  stdout = sys.stdout

  def handle_exception(self, e, message=""):
    if self.options["--verbose"]:
      print(e)
    if message:
      ColorPrint.red(message)

  def init(self):
    self.options["TARGET"] = self.options["TARGET"].split(",")
    # Clean up targets
    for i in range(len(self.options["TARGET"])):
      url = self.options["TARGET"][i]
      # Inject protocol if not there
      if not re.match(r'http(s?):', url):
        url = 'http://' + url

      parsed = urlsplit(url)
      host = parsed.netloc

      self.options["TARGET"][i] = host

      try:
        ColorPrint.green("Searching for subdomains for " + socket.gethostbyname(
          self.options["TARGET"][i]) + " (" + self.options["TARGET"][i] + ")")
      except Exception as e:
        self.handle_exception(e,
                              "Error connecting to target! Make sure you spelled it correctly and it is a resolvable address")
        raise e

  def run(self):
    # Retrieve IP of target and run initial configurations
    self.init()
    # If multiple targets, create scans for each
    for i in range(len(self.options["TARGET"])):
      # Default scans that run every time
      target = self.options["TARGET"][i]
      threads = [threading.Thread(target=dns_zonetransfer, args=(self, target)),
                 threading.Thread(target=subdomain_hackertarget,
                                  args=(self, target)),
                 threading.Thread(target=search_subject_alt_name,
                                  args=(self, target)),
                 threading.Thread(target=search_virustotal,
                                  args=(self, target)),
                 # threading.Thread(target=search_pkey, args=(self, target)),
                 # Removed pkey as of June 18 2018 due to issues on their end (not connecting)
                 threading.Thread(target=search_netcraft, args=(self, target)),
                 threading.Thread(target=search_crtsh, args=(self, target)),
                 threading.Thread(target=search_dnsdumpster,
                                  args=(self, target)),
                 threading.Thread(target=search_anubisdb, args=(self, target))]

      # Additional options - shodan.io scan
      if self.options["--additional-info"]:
        threads.append(threading.Thread(target=search_shodan, args=(self,)))

      # Additional options - ssl
      if self.options["--ssl"]:
        threads.append(threading.Thread(target=ssl_scan, args=(self, target)))

      # Additional options - nmap scan of dnssec script and a host/port scan
      if self.options["--with-nmap"]:
        threads.append(
          threading.Thread(target=dnssecc_subdomain_enum, args=(self, target)))
        threads.append(threading.Thread(target=scan_host, args=(self,)))

    # Start all threads and wait for them to finish
    for x in threads:
      x.start()

    for x in threads:
      x.join()

    # Run a recursive search on each subdomain - rarely useful, but nice to have
    # just in case
    if self.options["--recursive"]:
      recursive_search(self)

    # remove duplicates and clean up
    self.domains = self.clean_domains(self.domains)
    self.dedupe = set(self.domains)

    print("Found", len(self.dedupe), "subdomains")
    print("----------------")

    if self.options["--ip"]:
      self.resolve_ips()
    else:
      for domain in self.dedupe:
        ColorPrint.green(domain.strip())

    if self.options["--send-to-anubis-db"]:
      send_to_anubisdb(self, self.options["TARGET"])

  def resolve_ips(self):
    unique_ips = set()
    for domain in self.dedupe:
      try:
        # Attempt to get IP
        resolved_ip = socket.gethostbyname(domain)
      except Exception as e:
        self.handle_exception(e)
        # If getting IP fails, fallback to empty string
        resolved_ip = ""
      # TODO - Align domains and ips in stdout
      ColorPrint.green(domain + ": " + resolved_ip)
      if resolved_ip:
        unique_ips.add(resolved_ip)
    print("Found %s unique IPs" % len(unique_ips))
    for ip in unique_ips:
      # Ignore empty strings, final sanity check
      if ip:
        ColorPrint.green(ip)

  @staticmethod
  def clean_domains(domains):
    cleaned = []
    for subdomain in domains:
      subdomain = subdomain.lower()
      if subdomain.find("//") != -1:
        subdomain = subdomain[subdomain.find("//") + 2:]
      # Some pkey return instances like example.com. - remove the final .
      if subdomain.endswith('.'):
        subdomain = subdomain[:-1]
      # sometimes we'll get something like /www.example.com
      if subdomain[0] in ["\\", ".", "/", "#", "$", "%"]:
        subdomain = subdomain[1:]
      # If it's an email address, only take the domain part
      if "@" in subdomain:
        subdomain = subdomain.split("@")
        # If it's an actual email like mail@example.com, take example.com
        if len(subdomain) > 1:
          subdomain = subdomain[1]
        else:
          # If for some reason it's example.com@, take example.com
          subdomain = subdomain[0]

      cleaned.append(subdomain.strip())
    return cleaned



import censys

from anubis.utils.ColorPrint import ColorPrint


def search_censys(self, target):
  print("Searching Censys")
  try:
    from anubis.API import CENSYS_ID, CENSYS_SECRET
  except ImportError:
    ColorPrint.red(
      "To run a Censys scan, you must add your API keys to anubis/API.py")
    return
  if not CENSYS_SECRET or not CENSYS_ID:
    ColorPrint.red(
      "To run a Censys scan, you must add your API keys to anubis/API.py")
    return
  # Print certificate information for domains
  c = censys.certificates.CensysCertificates(CENSYS_ID, CENSYS_SECRET)
  for cert in c.search("." + target):
    print(cert)
